import os

from azureml.core.compute import AksCompute
from azureml.core.environment import Environment
from azureml.core.model import InferenceConfig, Model
from azureml.core.webservice import AciWebservice, AksWebservice, Webservice
from azureml.exceptions import ComputeTargetException
from pydantic import BaseModel

from azure_helper.logger import get_logger
from azure_helper.utils.aml_interface import AMLInterface

__here__ = os.path.dirname(__file__)

log = get_logger()


class DeploymentSettings(BaseModel):
    """_summary_

    Args:
        BaseModel (_type_): _description_
    """

    deployment_service_name: str
    cpu_cores: int = 1
    gpu_cores: int = 0
    memory_gb: int = 1
    enable_app_insights: bool = True


class DeployModel:
    def __init__(
        self,
        aml_interface: AMLInterface,
        aml_env_name: str,
        model_name: str,
        deployment_settings: DeploymentSettings,
    ) -> None:
        """_summary_

        Args:
            aml_interface (AMLInterface): _description_
            aml_env_name (str): _description_
            model_name (str): _description_
            deployment_settings (DeploymentSettings): _description_
        """

        self.aml_interface = aml_interface
        self.workspace = aml_interface.workspace

        self.aml_env_name = aml_env_name
        self.model_name = model_name
        self.deployment_settings = deployment_settings

    def get_inference_config(
        self,
    ) -> InferenceConfig:
        """_summary_

        Returns:
            _type_: _description_
        """

        aml_env = Environment.get(
            workspace=self.workspace,
            name=self.aml_env_name,
        )
        scoring_script_path = os.path.join(__here__, "score.py")
        return InferenceConfig(
            entry_script=scoring_script_path,
            environment=aml_env,
        )

    def deploy_aciservice(
        self,
        *args,
        **kwargs,
    ) -> Webservice:
        """_summary_"""

        inference_config = self.get_inference_config()

        aci_deployment = AciWebservice.deploy_configuration(
            *args,
            **kwargs,
            cpu_cores=self.deployment_settings.cpu_cores,
            memory_gb=self.deployment_settings.memory_gb,
            enable_app_insights=self.deployment_settings.enable_app_insights,
        )

        model = self.workspace.models.get(self.model_name)

        service = Model.deploy(
            self.workspace,
            self.deployment_settings.deployment_service_name,
            [model],
            inference_config,
            aci_deployment,
        )

        service.wait_for_deployment(show_output=True)
        log.info(service.state)
        log.info(service.scoring_uri)

    def deploy_aksservice(
        self,
        aks_cluster_name: str,
        *args,
        **kwargs,
    ) -> Webservice:
        """_summary_

        Args:
            aks_cluster_name (str): _description_
        """

        try:
            aks_target = AksCompute(self.workspace, name=aks_cluster_name)
            log.info(
                f"k8s cluster {aks_cluster_name} found in workspace {self.workspace}",
            )
        except ComputeTargetException:
            log.error(
                f"k8s cluster {aks_cluster_name} was not found in workspace {self.workspace}",
            )
            raise

        inference_config = self.get_inference_config()

        aks_deployment = AksWebservice.deploy_configuration(
            *args,
            **kwargs,
            cpu_cores=self.deployment_settings.cpu_cores,
            memory_gb=self.deployment_settings.memory_gb,
            enable_app_insights=self.deployment_settings.enable_app_insights,
        )

        model = self.workspace.models.get(self.model_name)

        service = Model.deploy(
            self.workspace,
            self.deployment_settings.deployment_service_name,
            [model],
            inference_config,
            aks_deployment,
            aks_target,
        )

        service.wait_for_deployment(show_output=True)
        log.info(service.state)
        log.info(service.scoring_uri)

    def update_service(
        self,
    ):
        """_summary_"""
        inference_config = self.get_inference_config()
        service = Webservice(
            name=self.deployment_settings.deployment_service_name,
            workspace=self.workspace,
        )
        model = self.workspace.models.get(self.model_name)
        service.update(models=[model], inference_config=inference_config)
        log.info(service.state)
        log.info(service.scoring_uri)


def main():
    # Retrieve vars from env
    workspace_name = os.environ["AML_WORKSPACE_NAME"]
    resource_group = os.environ["RESOURCE_GROUP"]
    subscription_id = os.environ["SUBSCRIPTION_ID"]

    spn_credentials = {
        "tenant_id": os.environ["TENANT_ID"],
        "service_principal_id": os.environ["SPN_ID"],
        "service_principal_password": os.environ["SPN_PASSWORD"],
    }

    aml_interface = AMLInterface(
        spn_credentials,
        subscription_id,
        workspace_name,
        resource_group,
    )
    webservices = aml_interface.workspace.webservices.keys()
    if DEPLOYMENT_SERVICE_NAME not in webservices:
        deploy_service(aml_interface)
    else:
        update_service(aml_interface)



"""init() and run(data) are both reserved functions with reserved variables needed to deploy a model, whether it is on ACI or AKS.

init() defines how to laod the model.

run(data) defines how to handle the datas that are fed to the model from the REST endpoint.
"""
# import json

# import joblib
# import numpy as np
# from azureml.core.model import Model

# from azure_helper.utils.const import MODEL_NAME


# def init():
#     global model
#     model_path = Model.get_model_path(MODEL_NAME)
#     model = joblib.load(model_path)


# def run(data):
#     try:
#         data = json.loads(data)
#         data = data["data"]
#         result = model.predict(np.array(data))
#         return result.tolist()
#     except Exception as e:
#         error = str(e)
#         return error



import os
from abc import ABC, abstractmethod
from pathlib import Path

from azureml.core import Dataset, Datastore, Run
from azureml.core.model import Model
from skl2onnx import __max_supported_opset__, convert_sklearn
from skl2onnx.common.data_types import FloatTensorType
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import f1_score

from azure_helper.logger import get_logger

__here__ = os.path.dirname(__file__)

log = get_logger()


class Train(ABC):
    @abstractmethod
    def get_df_from_datastore_path(self, datastore, datastore_path):
        log.info(f"Loading dataset {datastore_path} from datastore {datastore.name}")
        pass

    @abstractmethod
    def prepare_data(self):
        pass

    def train_model(self):
        log.info("Start training model.")
        pass

    def evaluate_model(self):
        log.info("Start evaluating model.")
        pass

    def save_model(self, model):
        log.info("Saving model to ONNX format.")
        pass

    def register_model(self, model_path):
        pass


class TrainingLoopExemple(Train):
    def __init__(
        self,
        run: Run,
        trainig_datastore: str,
        model_name: str,
        target_name: str,
        project_name: str,
    ):
        """_summary_

        Args:
            run (Run): _description_
            trainig_datastore (str): _description_
            model_name (str): _description_
            target_name (str): _description_
            project_name (str): _description_
        """

        self.run = run
        self.model_name = model_name
        self.target_name = target_name
        self.project_name = project_name

        self.workspace = run.experiment.workspace
        self.trainig_datastore = trainig_datastore
        self.datastore = Datastore.get(run.experiment.workspace, trainig_datastore)

    def get_df_from_datastore_path(self, datastore, datastore_path):
        """_summary_

        Args:
            datastore (_type_): _description_
            datastore_path (_type_): _description_

        Returns:
            _type_: _description_
        """
        # In our example we only have single files,
        # but these may be daily data dumps
        log.info(f"Loading dataset {datastore_path} from datastore {datastore.name}")
        datastore_path = [(datastore, datastore_path)]
        dataset = Dataset.Tabular.from_delimited_files(
            path=datastore_path,
        )
        return dataset.to_pandas_dataframe()

    def prepare_data(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        x_train = self.get_df_from_datastore_path(
            self.datastore,
            f"{self.project_name}/train/X_train.csv",
        )

        y_train = self.get_df_from_datastore_path(
            self.datastore,
            f"{self.project_name}/train/y_train.csv",
        )
        y_train = y_train[self.target_name]

        x_test = self.get_df_from_datastore_path(
            self.datastore,
            f"{self.project_name}/test/X_test.csv",
        )

        y_test = self.get_df_from_datastore_path(
            self.datastore,
            f"{self.project_name}/test/y_test.csv",
        )
        y_test = y_test[self.target_name]

        return x_train, y_train, x_test, y_test

    def train_model(self, x_train, y_train):
        """_summary_

        Args:
            x_train (_type_): _description_
            y_train (_type_): _description_

        Returns:
            _type_: _description_
        """
        log.info("Start training model.")
        model = LogisticRegression()
        model.fit(x_train, y_train)
        return model

    def evaluate_model(self, model, x_test, y_test):
        """_summary_

        Args:
            model (_type_): _description_
            x_test (_type_): _description_
            y_test (_type_): _description_
        """
        log.info("Start evaluating model.")
        y_pred = model.predict(x_test)
        model_f1_score = f1_score(y_test, y_pred)
        self.run.log("F1_Score", model_f1_score)

    def save_model(self, model):
        """_summary_

        Args:
            model (_type_): _description_

        Returns:
            _type_: _description_
        """
        log.info("Saving model to ONNX format.")
        output_dir = Path("outputs")
        output_dir.mkdir(parents=True, exist_ok=True)
        # output_dir = os.path.join(__here__, "outputs")
        # os.makedirs(output_dir, exist_ok=True)
        model_path = output_dir / Path("model.onnx")
        # model_path = os.path.join(output_dir, "model.pkl")

        initial_types = [("float_input", FloatTensorType([None, model.n_features_in_]))]
        model_onnx = convert_sklearn(
            model,
            initial_types=initial_types,
            target_opset=__max_supported_opset__,
        )

        # Save the model
        with open("outputs/model.onnx", "wb") as f:
            f.write(model_onnx.SerializeToString())
        # joblib.dump(model, model_path)
        log.info("Model saved.")
        return model_path

    def register_model(self, model_path):
        """_summary_

        Args:
            model_path (_type_): _description_
        """
        self.run.upload_file(str(model_path), "outputs/model.onnx")

        model = self.run.register_model(
            model_name=self.model_name,
            model_path="outputs/model.onnx",
            model_framework=Model.Framework.ONNX,
        )
        self.run.log("Model_ID", model.id)
        log.info(
            f"Model registered with following informations, name : {model.name}, id : {model.id}, version : {model.version}.",
        )



import os
from pathlib import Path
from typing import List

from azureml.core import Environment, Experiment, ScriptRunConfig
from azureml.core.runconfig import DockerConfiguration, RunConfiguration
from azureml.pipeline.core import Pipeline
from azureml.pipeline.steps import PythonScriptStep

from azure_helper.logger import get_logger
from azure_helper.utils.aml_interface import AMLInterface

log = get_logger()


class AMLExperiment:
    def __init__(
        self,
        aml_interface: AMLInterface,
        aml_compute_name: str,
        aml_compute_instance: str,
        env_name: str,
        experiment_name: str,
        training_script_path: str,
        clean_after_run: bool = True,
    ) -> None:
        """_summary_

        Args:
            aml_interface (AMLInterface): _description_
            aml_compute_name (str): _description_
            aml_compute_instance (str): _description_
            env_name (str): _description_
            experiment_name (str): _description_
            training_script_path (str): _description_
            clean_after_run (bool, optional): _description_. Defaults to True.
        """

        self.interface = aml_interface
        self.aml_compute_name = aml_compute_name
        self.aml_compute_instance = aml_compute_instance
        self.env_name = env_name
        self.experiment_name = experiment_name
        self.clean_after_run = clean_after_run
        self.training_script_path = training_script_path

    def generate_run_config(self) -> RunConfiguration:

        run_config = RunConfiguration()
        docker_config = DockerConfiguration(use_docker=True)
        run_config.docker = docker_config

        aml_run_env = Environment.get(
            self.interface.workspace,
            self.env_name,
        )

        run_config.environment = aml_run_env

        compute_target = self.interface.get_compute_target(
            self.aml_compute_name,
            self.aml_compute_instance,
        )
        run_config.target = compute_target

        return run_config

    def submit_pipeline(self, steps: List[PythonScriptStep]):
        """_summary_

        Args:
            steps (List[PythonScriptStep]): _description_

        https://github.com/Azure/MachineLearningNotebooks/blob/master/how-to-use-azureml/machine-learning-pipelines/intro-to-pipelines/aml-pipelines-getting-started.ipynb

        This should be a registered AZML Pipeline with the followings steps :

        * Download raw datas from one datastore
        * Transform those datas in "gold" datas and store them in another datastore
        * Use these gold datas to train a model
        * Evaluate that model
        * Save and version that model

        """
        experiment = Experiment(self.interface.workspace, self.experiment_name)
        # src_dir = __here__
        # src_dir = str(Path.cwd())

        compute_target = self.interface.get_compute_target(
            self.aml_compute_name,
            self.aml_compute_instance,
        )

        pipeline = Pipeline(workspace=self.interface.workspace, steps=steps)

        log.info("Submitting Run")
        run = experiment.submit(config=pipeline)
        run.wait_for_completion(show_output=True)
        log.info("Run completed.")

        if self.clean_after_run:
            log.info("Deleting compute instance.")
            compute_target.delete()

    def submit_run(self):
        """_summary_"""

        experiment = Experiment(self.interface.workspace, self.experiment_name)
        # src_dir = __here__
        src_dir = str(Path.cwd())

        docker_config = DockerConfiguration(use_docker=True)
        script = ScriptRunConfig(
            source_directory=src_dir,
            script=self.training_script_path,
            docker_runtime_config=docker_config,
        )

        compute_target = self.interface.get_compute_target(
            self.aml_compute_name,
            self.aml_compute_instance,
        )

        script.run_config.target = compute_target

        aml_run_env = Environment.get(
            self.interface.workspace,
            self.env_name,
        )
        script.run_config.environment = aml_run_env

        log.info("Submitting Run")
        run = experiment.submit(config=script)
        run.wait_for_completion(show_output=True)
        log.info(f"Run completed : {run.get_metrics()}")

        if self.clean_after_run:
            log.info("Deleting compute instance.")
            compute_target.delete()



from pathlib import Path

from azureml.core.environment import Environment
from pydantic import BaseModel

from azure_helper.logger import get_logger
from azure_helper.utils.aml_interface import AMLInterface

log = get_logger()


class EnvSpecs(BaseModel):
    flavor: str
    spec_file: Path


class AMLEnvironment:
    def __init__(self, dist_dir: Path, base_image: str) -> None:
        """Initialisation of the class.

        Args:
            dist_dir (Path): _description_
            base_image (str): _description_
        """
        self.dist_dir = dist_dir
        self.base_image = base_image

    def validate_dir(self):
        """_summary_

        Raises:
            FileNotFoundError: _description_
        """
        if self.dist_dir.is_dir():
            log.info(f"Looking for wheel file in {self.dist_dir}.")
        else:
            raise FileNotFoundError

    def retrieve_whl_filepath(self):
        """_summary_

        Raises:
            FileNotFoundError: _description_

        Returns:
            _type_: _description_
        """
        try:
            self.validate_dir()
        except FileNotFoundError:
            log.error(f"Couldn't find distribution directory {self.dist_dir}")
            raise

        whl_file = sorted(
            Path(file)
            for file in Path(self.dist_dir).glob("**/*.whl")
            if file.is_file()
        )
        if len(whl_file) == 0:
            log.error("Couldn't find wheel distribution")
            raise FileNotFoundError

        log.info(f"Found wheel {self.dist_dir / Path(whl_file[0])}")

        return self.dist_dir / Path(whl_file[0])

    def create_aml_environment(
        self,
        env_name: str,
        env_specs: EnvSpecs,
        aml_interface: AMLInterface,
    ) -> Environment:
        """_summary_

        Args:
            env_name (str): _description_
            env_specs (EnvSpecs): _description_
            aml_interface (AMLInterface): _description_

        Raises:
            ValueError: _description_

        Returns:
            Environment: _description_
        """
        if env_specs.flavor == "pip":
            env = Environment.from_pip_requirements(
                name=env_name,
                file_path=str(env_specs.spec_file),
            )
        elif env_specs.flavor == "conda":
            env = Environment.from_conda_specification(
                name=env_name,
                file_path=str(env_specs.spec_file),
            )
        elif env_specs.flavor == "docker":
            env = Environment.from_dockerfile(
                name=env_name,
                dockerfile=str(env_specs.spec_file),
            )
        else:
            log.error(
                f"env_specs flavor {env_specs.flavor} is not a valid one. Only 'pip', 'conda', or 'docker' are valid choices.",
            )
            raise ValueError

        whl_filepath = self.retrieve_whl_filepath()
        private_wheel = env.add_private_pip_wheel(
            workspace=aml_interface.workspace,
            file_path=whl_filepath,
            exist_ok=True,
        )
        env.python.conda_dependencies.add_pip_package(private_wheel)

        # conda_dep.add_pip_package(whl_url)
        # https://docs.microsoft.com/en-us/azure/machine-learning/how-to-train-with-custom-image#set-up-a-training-experiment
        env.docker.base_image = self.base_image

        # env.python.user_managed_dependencies = True

        # https://stackoverflow.com/questions/67387249/how-to-use-azureml-core-runconfig-dockerconfiguration-class-in-azureml-core-envi

        return env



import json
import os
from unittest import TestCase
from unittest.mock import Mock, patch

import pandas as pd
from sklearn.linear_model import LogisticRegression

from azure_helper.steps.score import init, run
from azure_helper.utils.transform_data import remove_collinear_cols

__here__ = os.path.dirname(__file__)
test_data_dir = os.path.join(__here__, "test_data")


class TestInit(TestCase):
    @patch("src.my_custom_package.score.Model")
    def test_init(self, mock_model):
        mock_model.get_model_path.return_value = os.path.join(
            test_data_dir,
            "test_model.pkl",
        )

        init()
        mock_model.get_model_path.assert_called_once()

        # model has now been initialised
        from azure_helper.steps.score import model

        self.assertIsInstance(model, LogisticRegression)


class TestRun(TestCase):
    def setUp(self):
        X_data = pd.read_csv(os.path.join(test_data_dir, "test_X_valid.csv"))
        X_data = remove_collinear_cols(X_data)
        self.test_data = json.dumps({"data": X_data.values.tolist()})

    def test_run(self):
        result = run(self.test_data)
        self.assertIsInstance(result, list)
        for y_value in result:
            self.assertIn(y_value, [0, 1])

    def test_run_error(self):
        result = run('{"data": [0, 1, 0]}')
        self.assertIn("Expected 2D array, got 1D array instead", result)



import pandas as pd

from azure_helper.utils.blob_storage_interface import BlobStorageInterface


class CreateData:
    def __init__(
        self,
        project_name: str,
        train_datastore: str = "train",
        test_datastore: str = "test",
    ):
        """_summary_

        Args:
            project_name (str): _description_
            train_datastore (str, optional): _description_. Defaults to "train".
            test_datastore (str, optional): _description_. Defaults to "test".
        """
        self.project_name = project_name
        self.train_datastore = train_datastore
        self.test_datastore = test_datastore

    def upload_training_data(
        self,
        blob_storage_interface: BlobStorageInterface,
        x_train: pd.DataFrame,
        y_train: pd.DataFrame,
    ):
        """_summary_

        Args:
            blob_storage_interface (BlobStorageInterface): _description_
            x_train (pd.DataFrame): _description_
            y_train (pd.DataFrame): _description_
        """
        blob_storage_interface.upload_df_to_blob(
            dataframe=x_train,
            container_name=f"{self.project_name}",
            blob_path=f"{self.train_datastore}/X_train.csv",
        )
        blob_storage_interface.upload_df_to_blob(
            dataframe=y_train,
            container_name=f"{self.project_name}",
            blob_path=f"{self.train_datastore}/y_train.csv",
        )

    def upload_validation_data(
        self,
        blob_storage_interface: BlobStorageInterface,
        x_valid: pd.DataFrame,
        y_valid: pd.DataFrame,
    ):
        """_summary_

        Args:
            blob_storage_interface (BlobStorageInterface): _description_
            x_valid (pd.DataFrame): _description_
            y_valid (pd.DataFrame): _description_
        """
        # Data to be used during model validation
        blob_storage_interface.upload_df_to_blob(
            dataframe=x_valid,
            container_name=f"{self.project_name}",
            blob_path=f"{self.train_datastore}/X_valid.csv",
        )
        blob_storage_interface.upload_df_to_blob(
            dataframe=y_valid,
            container_name=f"{self.project_name}",
            blob_path=f"{self.train_datastore}/y_valid.csv",
        )

    def upload_test_data(
        self,
        blob_storage_interface: BlobStorageInterface,
        x_test: pd.DataFrame,
        y_test: pd.DataFrame,
    ):
        """_summary_

        Args:
            blob_storage_interface (BlobStorageInterface): _description_
            x_test (pd.DataFrame): _description_
            y_test (pd.DataFrame): _description_
        """
        # Data to be used during model evaluation
        # So stored in the training container
        blob_storage_interface.upload_df_to_blob(
            dataframe=x_test,
            container_name=f"{self.project_name}",
            blob_path=f"{self.test_datastore}/X_test.csv",
        )
        blob_storage_interface.upload_df_to_blob(
            dataframe=y_test,
            container_name=f"{self.project_name}",
            blob_path=f"{self.test_datastore}/y_test.csv",
        )



import json
import os

import numpy as np
import requests
from azureml.core.webservice import Webservice
from sklearn.metrics import f1_score

from azure_helper.utils.aml_interface import AMLInterface
from azure_helper.utils.blob_storage_interface import BlobStorageInterface
from azure_helper.utils.const import DEPLOYMENT_SERVICE_NAME, SCORING_CONTAINER
from azure_helper.utils.transform_data import remove_collinear_cols


def get_validation_data(storage_acct_name, storage_acct_key):
    blob_storage_interface = BlobStorageInterface(
        storage_acct_name,
        storage_acct_key,
    )
    x_valid = blob_storage_interface.download_blob_to_df(
        container_name=SCORING_CONTAINER,
        blob_path="X_valid.csv",
    )
    y_valid = blob_storage_interface.download_blob_to_df(
        container_name=SCORING_CONTAINER,
        blob_path="y_valid.csv",
    )
    return x_valid, y_valid


def get_web_service_uri(aml_interface):
    service = Webservice(
        name=DEPLOYMENT_SERVICE_NAME,
        workspace=aml_interface.workspace,
    )
    return service.scoring_uri


def make_predictions(x_df, scoring_uri):
    data = json.dumps({"data": x_df.values.tolist()})
    headers = {"Content-Type": "application/json"}
    response = requests.post(scoring_uri, data=data, headers=headers)
    return np.array(response.json())


def score_predictions(y_valid, y_pred):
    model_validation_f1_score = round(f1_score(y_valid, y_pred), 3)
    print(f"F1 Score on Validation Data Set: {model_validation_f1_score}")


def main():
    storage_acct_name = os.environ["STORAGE_ACCT_NAME"]
    storage_acct_key = os.environ["STORAGE_ACCT_KEY"]
    workspace_name = os.environ["AML_WORKSPACE_NAME"]
    resource_group = os.environ["RESOURCE_GROUP"]
    subscription_id = os.environ["SUBSCRIPTION_ID"]

    spn_credentials = {
        "tenant_id": os.environ["TENANT_ID"],
        "service_principal_id": os.environ["SPN_ID"],
        "service_principal_password": os.environ["SPN_PASSWORD"],
    }

    aml_interface = AMLInterface(
        spn_credentials,
        subscription_id,
        workspace_name,
        resource_group,
    )
    x_valid, y_valid = get_validation_data(storage_acct_name, storage_acct_key)
    x_valid = remove_collinear_cols(x_valid)
    y_valid = y_valid["Target"]
    aml_interface = AMLInterface(
        spn_credentials,
        subscription_id,
        workspace_name,
        resource_group,
    )
    scoring_uri = get_web_service_uri(aml_interface)
    y_pred = make_predictions(x_valid, scoring_uri)
    score_predictions(y_valid, y_pred)


if __name__ == "__main__":
    main()



import os
import shutil
from unittest import TestCase
from unittest.mock import Mock, patch

import joblib
import pandas as pd
from sklearn.linear_model import LogisticRegression

from azure_helper.steps.train import (
    evaluate_model,
    get_df_from_datastore_path,
    prepare_data,
    register_model,
    save_model,
    train_model,
)

__here__ = os.path.dirname(__file__)
test_data_dir = os.path.join(__here__, "test_data")


@patch("src.my_custom_package.train.Dataset")
def test_get_df_from_datastore_path(mock_dataset):
    mock_datastore = Mock()
    datastore_path = "test_path"
    mock_dataset_obj = Mock()
    mock_dataset.Tabular.from_delimited_files.return_value = mock_dataset_obj
    test_df = pd.DataFrame([{"a": 1, "b": 2}, {"a": 3, "b": 4}])
    mock_dataset_obj.to_pandas_dataframe.return_value = test_df

    df = get_df_from_datastore_path(mock_datastore, datastore_path)
    mock_dataset.Tabular.from_delimited_files.assert_called_once_with(
        path=[(mock_datastore, datastore_path)],
    )
    mock_dataset_obj.to_pandas_dataframe.assert_called_once()
    assert df.equals(test_df)


class TestPrepareData(TestCase):
    @patch("src.my_custom_package.train.get_df_from_datastore_path")
    @patch("src.my_custom_package.train.Datastore")
    def test_get_df_from_datastore_path(
        self,
        mock_datastore,
        mock_get_df_from_datastore_path,
    ):
        # get_df_from_datastore_path tested above
        mock_get_df_from_datastore_path.side_effect = [
            pd.read_csv(os.path.join(test_data_dir, "test_X_train.csv")),
            pd.read_csv(os.path.join(test_data_dir, "test_y_train.csv")),
            pd.read_csv(os.path.join(test_data_dir, "test_X_test.csv")),
            pd.read_csv(os.path.join(test_data_dir, "test_y_test.csv")),
        ]
        X_train, y_train, X_test, y_test = prepare_data("test_workspace")

        self.assertIsInstance(X_train, pd.DataFrame)
        self.assertIsInstance(y_train, pd.Series)
        self.assertIsInstance(X_test, pd.DataFrame)
        self.assertIsInstance(y_test, pd.Series)

        self.assertIn("A", X_train.columns)
        self.assertIn("C", X_train.columns)
        self.assertIn("J", X_train.columns)
        self.assertIn("C", X_test.columns)
        self.assertNotIn("D", X_train.columns)
        self.assertNotIn("I", X_train.columns)
        self.assertNotIn("D", X_test.columns)
        self.assertNotIn("I", X_test.columns)

        self.assertEqual(len(X_train), len(y_train))
        self.assertEqual(len(X_test), len(y_test))


class TestTrainModel(TestCase):
    @patch("src.my_custom_package.train.get_df_from_datastore_path")
    @patch("src.my_custom_package.train.Datastore")
    def setUp(self, mock_datastore, mock_get_df_from_datastore_path):
        # prepare_data tested above
        mock_get_df_from_datastore_path.side_effect = [
            pd.read_csv(os.path.join(test_data_dir, "test_X_train.csv")),
            pd.read_csv(os.path.join(test_data_dir, "test_y_train.csv")),
            pd.read_csv(os.path.join(test_data_dir, "test_X_test.csv")),
            pd.read_csv(os.path.join(test_data_dir, "test_y_test.csv")),
        ]
        self.X_train, self.y_train, _, _ = prepare_data("test_workspace")

    def test_train_model(self):
        trained_model = train_model(self.X_train, self.y_train)
        self.assertIsInstance(trained_model, LogisticRegression)


class TestEvaluateModel(TestCase):
    @patch("src.my_custom_package.train.get_df_from_datastore_path")
    @patch("src.my_custom_package.train.Datastore")
    def setUp(self, mock_datastore, mock_get_df_from_datastore_path):
        # prepare_data tested above
        mock_get_df_from_datastore_path.side_effect = [
            pd.read_csv(os.path.join(test_data_dir, "test_X_train.csv")),
            pd.read_csv(os.path.join(test_data_dir, "test_y_train.csv")),
            pd.read_csv(os.path.join(test_data_dir, "test_X_test.csv")),
            pd.read_csv(os.path.join(test_data_dir, "test_y_test.csv")),
        ]
        _, _, self.X_test, self.y_test = prepare_data("test_workspace")
        self.classifier = joblib.load(
            os.path.join(test_data_dir, "test_model.pkl"),
        )

    def test_evaluate_model(self):
        mock_run = Mock()

        evaluate_model(self.classifier, self.X_test, self.y_test, mock_run)

        mock_run.log.assert_called_once()
        # First argument of first call
        self.assertEqual(mock_run.log.call_args[0][0], "F1_Score")
        # Second argument of first call
        self.assertIsInstance(mock_run.log.call_args[0][1], float)


class TestSaveModel(TestCase):
    def setUp(self):
        self.classifier = joblib.load(
            os.path.join(test_data_dir, "test_model.pkl"),
        )

    @patch("src.my_custom_package.train.__here__", test_data_dir)
    def test_save_model(self):
        save_model(self.classifier)
        self.assertTrue(
            os.path.exists(
                os.path.join(test_data_dir, "outputs"),
            ),
        )
        self.assertTrue(
            os.path.exists(
                os.path.join(test_data_dir, "outputs", "model.pkl"),
            ),
        )

    def tearDown(self):
        outputs_dir = os.path.join(test_data_dir, "outputs")
        shutil.rmtree(outputs_dir, ignore_errors=True)


def test_register_model():
    mock_run = Mock()
    mock_azure_model_obj = Mock()
    mock_azure_model_obj.id = "test_id"
    mock_run.register_model.return_value = mock_azure_model_obj

    register_model(mock_run, "test_model_path")
    mock_run.upload_file.assert_called_once_with(
        "test_model_path",
        "outputs/model.pkl",
    )
    mock_run.register_model.assert_called_once()
    mock_run.log.assert_called_once_with("Model_ID", "test_id")



from unittest.mock import Mock

from azureml.exceptions import ComputeTargetException
from pytest import fixture

from azure_helper.utils.aml_interface import AMLInterface

test_module = "azure_helper.utils.aml_interface"


@fixture
def aml_interface(mocker):
    mocker_spn = mocker.patch(
        f"{test_module}.ServicePrincipalAuthentication",
    )
    mock_workspace = mocker.patch(
        f"{test_module}.Workspace",
        return_value="test_workspace",
    )
    spn_credentials = {
        "tenant_id": "test_tenant_id",
        "service_principal_id": "test_spn_id",
        "service_principal_password": "test_spn_password",
    }
    return AMLInterface(
        spn_credentials=spn_credentials,
        subscription_id="test_subscription_id",
        workspace_name="test_workspace_name",
        resource_group="test_resource_group",
    )


class TestAMLInterface:
    def test_register_datastore(self, mocker, aml_interface):

        mock_datastore = mocker.patch(
            f"{test_module}.Datastore",
        )

        aml_interface.register_datastore(
            "test_datastore_name",
            "test_blob_container",
            "test_storage_acct_name",
            "test_storage_acct_key",
        )

        mock_datastore.register_azure_blob_container.assert_called_once_with(
            workspace="test_workspace",
            datastore_name="test_datastore_name",
            container_name="test_blob_container",
            account_name="test_storage_acct_name",
            account_key="test_storage_acct_key",
        )

    def register_aml_environment(self, aml_interface):
        mock_environment = Mock()

        aml_interface.register_aml_environment(mock_environment)

        mock_environment.register.assert_called_once_with("test_workspace")

    def test_get_compute_target(self, mocker, aml_interface):
        mock_compute_target = mocker.patch(
            f"{test_module}.ComputeTarget",
        )
        mock_AmlCompute = mocker.patch(
            f"{test_module}.AmlCompute",
        )

        mock_compute_target_obj = Mock()
        mock_compute_target.create.return_value = mock_compute_target_obj

        # First call to mock_compute_target returns 'test_compute_target'
        compute_target = aml_interface.get_compute_target(
            "test_compute_name",
            "STANDARD_D2_V2",
        )
        #  {'currentNodeCount': 1, 'targetNodeCount': 1, 'nodeStateCounts': {'preparingNodeCount': 0, 'runningNodeCount': 0, 'idleNodeCount': 1, 'unusableNodeCount': 0, 'leavingNodeCount': 0, 'preemptedNodeCount': 0}, 'allocationState': 'Steady', 'allocationStateTransitionTime': '2022-08-11T13:48:24.409000+00:00', 'errors': None, 'creationTime': '2022-08-11T13:46:01.432646+00:00', 'modifiedTime': '2022-08-11T13:46:08.602275+00:00', 'provisioningState': 'Succeeded', 'provisioningStateTransitionTime': None, 'scaleSettings': {'minNodeCount': 1, 'maxNodeCount': 2, 'nodeIdleTimeBeforeScaleDown': 'PT1800S'}, 'vmPriority': 'Dedicated', 'vmSize': 'STANDARD_D2_V2'}
        # assert mock_compute_target_obj.status.serialized().vmSize == "STANDARD_D2_V2"

        mock_compute_target.create.assert_not_called()

        mock_compute_target.side_effect = ComputeTargetException(
            "Compute Target Not Found",
        )
        mock_compute = Mock()
        # Compute target exists, create not called
        mock_compute_target.create.return_value = mock_compute

        # Second call to mock_compute_target raises ComputeTargetException
        # Suggesting the compute target needs to be created
        output_2 = aml_interface.get_compute_target(
            "test_compute_name",
            "STANDARD_D2_V2",
        )

        assert output_2 == mock_compute

        mock_AmlCompute.provisioning_configuration.assert_called_once_with(
            vm_size="STANDARD_D2_V2",
            min_nodes=1,
            max_nodes=2,
        )
        mock_compute_target.create.assert_called_once()
        mock_compute.wait_for_completion.assert_called_once()



from io import StringIO

import pandas as pd
from azure.core.exceptions import ResourceExistsError
from azure.storage.blob import BlobServiceClient

from azure_helper.logger import get_logger

log = get_logger()


class BlobStorageInterface:
    def __init__(self, storage_acct_name: str, storage_acct_key: str):
        """_summary_

        Args:
            storage_acct_name (str): _description_
            storage_acct_key (str): _description_
        """
        conn_str = (
            "DefaultEndpointsProtocol=https;"
            + f"AccountName={storage_acct_name};"
            + f"AccountKey={storage_acct_key};"
            + "EndpointSuffix=core.windows.net"
        )
        self.blob_service_client = BlobServiceClient.from_connection_string(
            conn_str,
        )

    def create_container(self, container_name: str):
        """_summary_

        Args:
            container_name (str): _description_
        """
        try:
            self.blob_service_client.create_container(container_name)
            log.info(f"Creating blob storage container {container_name}.")
        except ResourceExistsError:
            log.warning(f"Blob storage container {container_name} already exists.")
            pass

    def upload_df_to_blob(
        self,
        dataframe: pd.DataFrame,
        container_name: str,
        blob_path: str,
    ):
        """_summary_

        Args:
            dataframe (pd.DataFrame): _description_
            container_name (str): _description_
            blob_path (str): _description_
        """
        self.create_container(container_name)
        blob_client = self.blob_service_client.get_blob_client(
            container=container_name,
            blob=blob_path,
        )
        try:
            blob_client.upload_blob(
                dataframe.to_csv(index=False, header=True).encode(),
            )
            log.info(f"Dataset uploaded at blob path : {blob_path}.")
        except ResourceExistsError:
            log.warning(
                f"Blob path {blob_path} already contains datas. Now deleting old datas tu upload the new ones.",
            )
            blob_client.delete_blob()
            blob_client.upload_blob(
                dataframe.to_csv(index=False, header=True).encode(),
            )
            log.info(f"New dataset uploaded at blob path : {blob_path}.")

    def download_blob_to_df(self, container_name: str, blob_path: str):
        """_summary_

        Args:
            container_name (str): _description_
            blob_path (str): _description_

        Returns:
            _type_: _description_
        """
        blob_client = self.blob_service_client.get_blob_client(
            container=container_name,
            blob=blob_path,
        )
        stream = blob_client.download_blob()
        buffer = StringIO(stream.content_as_text())
        dataframe = pd.read_csv(buffer)
        log.info(f"Download from {container_name} ended successfully.")
        return dataframe
